// @TITLE "dirichlet.cc -- parallel Dirichlet program"
//
// dirichlet -- a parallel Dirichlet relaxation program
//   written using the Data-Parallel Programming Library for Education (DAPPLE)
//
// The Dirichlet function is really just a 4-nearest-neighbors
// averaging stencil, applied repeatedly over many iterations
// to a matrix whose boundary conditions are fixed; eventually,
// the values converge to a steady state.  In this example we 
// pick some arbitrary boundary values so that two opposite corners 
// are high, the other two corners are low, and the other boundary 
// values are linearly interpolated between the corners.
//
// Note that we precompute a boolean vector that remembers which
// points are internal, to simplify the loop.  Also, note the use of
// a MAX reduction to determine convergence.
//
// David Kotz 1994

// $Id: dirichlet.cc,v 1.9 95/12/17 00:08:30 dfk release Locker: dfk $

#include <stdlib.h>
#include <iostream.h>
#include <math.h>
#include "dapple.h"

// we'll work with NxN matrices
const int N = 5;
const float epsilon = 0.05;

// which row or column am I? useful for VP-specific math
const intMatrix VProw(N, N, IdentityR);
const intMatrix VPcol(N, N, IdentityC);

int
main()
{
    floatMatrix M(N,N);		   // current values of matrix
    floatMatrix Mprev(N,N);	   // previous values of matrix
    booleanMatrix isInternal(N,N); // false only on the edges of the matrix
    int iteration = 0;

    // initialize the matrix
    // internal points should be zero, but boundary points could be anything
    // I chose something simple.  Note the corners are initialized with the
    // top and bottom rows, although the function I've chosen for the boundary 
    // values is consistent at the corners.
    ifp (VProw == 0) {          // top boundary
	isInternal = false;
	M = VPcol + 1;		// 1..N
    } else ifp (VProw == N-1) {	// bottom boundary
	isInternal = false;
	M = N - VPcol;		// N..1
    } else ifp (VPcol == 0) {	// left boundary
	isInternal = false;
	M = VProw + 1;		// 1..N
    } else ifp (VPcol == N-1) { // right boundary
	isInternal = false;
	M = N - VProw;		// N..1
    } else {			// internal points
	isInternal = true;
	M = 0;
    }

    cout << "initial state:" << endl;
    cout << M << endl;

    // iterate until the maximum change is epsilon or less
    do {
	Mprev = M;		// remember previous value
	ifp (isInternal) {
	    M = (  shift(M, 0, -1)	// right neighbor
		 + shift(M, 0, 1)	// left neighbor
		 + shift(M, -1, 0)	// top neighbor
		 + shift(M, 1, 0));	// bottom neighbor
	    M = M / 4;
	}

	cout << "after iteration " << ++iteration << ":" << endl;
	cout << M << endl;
	
    } while (max_value(M - Mprev) > epsilon);

    return(0);
}
