// @TITLE "nbody.cc -- parallel nbody solution"
//
// nbody -- a parallel  program to solve the n-bodies problem (in 2 dimensions)
//   written using the Data-Parallel Programming Library for Education (DAPPLE)
//
// This program has lots of parallelism.  It is an n^2 algorithm, in that
// each body must compute the force on itself caused by all the other n bodies.
// So, we loop over the n nodies, and in each iteration we compute the
// forces caused by the other bodies, in parallel.  Then, we update the
// position and velocity of all n bodies in parallel, using the forces
// from the previous step.  Over all this we iterate through timesteps.
//
//    The interesting parallel parts are in ComputeForces and MoveBodies.
//
// David Kotz 1994

// $Id: nbody.cc,v 1.7 95/12/17 00:08:09 dfk release Locker: dfk $

#include <stdlib.h>
#include <iostream.h>
#include <math.h>
#include "dapple.h"

void ComputeForces(const int N, const doubleVector& Mass, 
		   const doubleVector& Xpos, const doubleVector& Ypos, 
		   const doubleVector& Xvel, const doubleVector& Yvel, 
		   doubleVector& Xforce, doubleVector& Yforce);
void MoveBodies(const int N, const doubleVector& Mass, 
		doubleVector& Xpos, doubleVector& Ypos, 
		doubleVector& Xvel, doubleVector& Yvel, 
		const doubleVector& Xforce, const doubleVector& Yforce,
		double& time, const double delta_t);
void PrintBodies(const int N, const doubleVector& Mass, 
		 const doubleVector& Xpos, const doubleVector& Ypos, 
		 const doubleVector& Xvel, const doubleVector& Yvel,
		 const double time);
boolean ReadBodies(FILE *fp, const int N, const doubleVector& Mass, 
		   const doubleVector& Xpos, const doubleVector& Ypos, 
		   const doubleVector& Xvel, const doubleVector& Yvel);

/////////////////////////////////////////////////////////////
// @SUBTITLE "Main program"
int
main(int argc, char **argv)
{
    char *filename;			// name of input file
    double delta_t;			// timestep
    int steps;				// number of timesteps
    int N;				// number of bodies

    if (argc != 4) {
	fprintf(stderr, "usage: nbody filename delta_t #steps\n");
	exit(1);
    }

    filename = argv[1];

    delta_t = atof(argv[2]);
    if (delta_t <= 0) {
	fprintf(stderr, "nbody: timestep (%f) should be > 0\n", delta_t);
	exit(-1);
    }

    steps = atoi(argv[3]);
    if (steps <= 0) {
	fprintf(stderr, "nbody: number of timesteps (%d) should be > 0\n", steps);
	exit(-1);
    }

    // open the file
    FILE *fp;
    if ((fp = fopen(filename, "r")) == (FILE *)NULL) {
	fprintf(stderr, "nbody: cannot open input file '%s'\n", filename);
	exit(1);
    }

    // read the number of bodies
    if (fscanf(fp, "%d\n", &N) != 1) {
	fprintf(stderr, "nbody: file '%s': missing value of N on first line\n",
		filename);
	fclose(fp);
	exit(1);
    }

    // allocate vectors to represent the state of N bodies
    doubleVector Mass(N);		  // mass in kilograms
    doubleVector Xpos(N), Ypos(N);	  // x,y position in meters
    doubleVector Xvel(N), Yvel(N);	  // x,y velocity in meters/second
    doubleVector Xforce(N), Yforce(N);	  // x,y force components in Newtons
    double time = 0;			  // the current time in seconds
    
    // read in the N bodies. 
    if (! ReadBodies(fp, N, Mass, Xpos, Ypos, Xvel, Yvel)) {
	fclose(fp);
	exit(1);
    } else
      fclose(fp);

    ////////////////////////////////////////////////////
    // OK, time to do the simulation
    for (int i = 0; i < steps; i++) {
	ComputeForces(N, Mass, Xpos, Ypos, Xvel, Yvel, Xforce, Yforce);
	MoveBodies(N, Mass, Xpos, Ypos, Xvel, Yvel, Xforce, Yforce, time, delta_t);
	PrintBodies(N, Mass, Xpos, Ypos, Xvel, Yvel, time);
    }

    return(0);
}

/////////////////////////////////////////////////////////////
// @SUBTITLE "ReadBodies: read an input file"
// Data file format:
//  N
//  body 0
//  body 1
//  body 2
//   ...
//  body N-1
//
// where each body is 5 floats on a line, separated by spaces:
//         mass, x,  y, x velocity, y velocity
// units:   kg   m  m     m/s         m/s
//
boolean	// true if success
ReadBodies(FILE *fp, const int N, const doubleVector& Mass, 
	   const doubleVector& Xpos, const doubleVector& Ypos, 
	   const doubleVector& Xvel, const doubleVector& Yvel)
{
    for (int i = 0; i < N; i++) {
	// 5 floats for each body: mass, x, y, x velocity, y velocity
	if (fscanf(fp, "%lf %lf %lf %lf %lf\n", 
		   &Mass[i], &Xpos[i], &Ypos[i], &Xvel[i], &Yvel[i]) != 5) {
	    fprintf(stderr, "nbody: bad or missing data for body %d\n", i);
	    return(false);
	}
	if (Mass[i] <= 0) {
	    fprintf(stderr, "nbody: zero or negative mass for body %d\n", i);
	    return(false);
	}
    }
    return(true);
}

/////////////////////////////////////////////////////////////
// @SUBTITLE "ComputeForces: compute the total force on each body"
// The total force on each body is the sum of the forces 
// on it by all other bodies.
// We compute the force for each body sequentially, but we can
// compute, in parallel, the force on it by all other bodies.

void ComputeForces(const int N, const doubleVector& Mass, 
		   const doubleVector& Xpos, const doubleVector& Ypos, 
		   const doubleVector& Xvel, const doubleVector& Yvel, 
		   doubleVector& Xforce, doubleVector& Yforce)
{
    // distance between one body and each other body
    doubleVector dx(N), dy(N), d(N); // x and y components, and total distance
    doubleVector f(N);		     // scalar component of force
    const intVector VP(N, Identity); // 0, 1, 2, ..., N-1
    // gravitational constant G: negative since it is attractive
    // Units: Nm^2/kg^2; thus force is in Newtons
    const double GRAVITY = -6.670e-11;

    for (int body = 0; body < N; body++)
      ifp (VP != body) {
	  // find the distance to all other bodies
	  dx = Xpos[body] - Xpos;
	  dy = Ypos[body] - Ypos;
	  d = apply(sqrt, dx*dx + dy*dy);
	  
	  // find the scalar component of the force caused by all other bodies
	  f = GRAVITY * Mass[body] * Mass / (d * d * d);
	  
	  // sum all the forces in both the x and y directions
	  Xforce[body] = sum(dx * f);
	  Yforce[body] = sum(dy * f);
      }
}

/////////////////////////////////////////////////////////////
// @SUBTITLE "MoveBodies: update the position/velocity from accumulated force"
void MoveBodies(const int N, const doubleVector& Mass, 
		doubleVector& Xpos, doubleVector& Ypos, 
		doubleVector& Xvel, doubleVector& Yvel, 
		const doubleVector& Xforce, const doubleVector& Yforce,
		double& time, const double delta_t)
{
    doubleVector ax(N), ay(N);	// the acceleration due to the force

    // a = F/m
    ax = Xforce / Mass;
    ay = Yforce / Mass;

    // x' = x + vt + 1/2 at^2
    Xpos = Xpos + Xvel * delta_t + 0.5 * ax * delta_t * delta_t;
    Ypos = Ypos + Yvel * delta_t + 0.5 * ay * delta_t * delta_t;

    // v' = v + at
    Xvel = Xvel + ax * delta_t;
    Yvel = Yvel + ay * delta_t;

    // t' = t + dt
    time += delta_t;
}

/////////////////////////////////////////////////////////////
// @SUBTITLE "PrintBodies: print the state of all bodies"
void PrintBodies(const int N, const doubleVector& Mass, 
		 const doubleVector& Xpos, const doubleVector& Ypos, 
		 const doubleVector& Xvel, const doubleVector& Yvel,
		 const double time)
{
    printf("State of the %d bodies at time %g seconds:\n", N, time);
    printf("%4s %12s %12s %12s %12s %12s\n",
	   "Body", "Mass", "X", "Y", "Xvel", "Yvel");

    for (int body = 0; body < N; body++)
      printf("%4d %12g %12g %12g %12g %12g\n", 
	     body, Mass[body], Xpos[body], Ypos[body], Xvel[body], Yvel[body]);
    printf("\n");
    fflush(stdout);
}

