// @TITLE "Collection.cc: class Collection"
//
//  Collection class; included by {int,char,float,double,boolean}.cc
//
// David Kotz 1994
//
// initially derived from 
//    ``A C++ Primer''  by Stanley Lippman, Addison-Wesley
//     Copyright (c) 1991 AT&T Bell Laboratories, All Rights Reserved   

// $Id: Collection.cc,v 2.16 95/12/17 00:07:55 dfk release Locker: dfk $

#if !defined(BOOLEAN_ELEMENTS) && !defined(INT_ELEMENTS)
// max and min functions for elements; note this overloads int version in macros.h 
static inline Element max(CONST Element x, CONST Element y) { return (x > y ? x : y); }
static inline Element min(CONST Element x, CONST Element y) { return (x < y ? x : y); }
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::construct
// called by all constructors, to get the basic stuff set up

void Collection::construct(void)
{
    number = shape.number();
    v = new Element[number]; 

    INVARIANT3(v != NULL, 
	       "Out of memory for making new collection of %d elements.", 
	       number);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::init: private function to initialize collection"
// This, or the next one, are called by most of the constructors.
// Initialize all values to a scalar value x.
// Note that all entries are initialized, regardless of active set.

void Collection::init(CONST Element x)
{
    // now initialize v[] using the scalar
    for (int i = 0; i < number; i++)
      (*this)(i) = x;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Copy and conversion constructors"
// These are a set of constructors that can build one collection
// from another, possibly of a different elemental type.  
// Initialize all values as a copy of the other collection C.
// Thus, these roll the constructor and init functions together.
// Note that all entries are initialized, regardless of active set.

Collection::Collection(CONST intCollection& C) : shape(C.shape)
{
    construct(); 

    // now initialize v[] using the array of C
    for (int i = 0; i < number; i++)
      (*this)(i) = Element(C(i));
}

Collection::Collection(CONST charCollection& C) : shape(C.shape)
{ 
    construct(); 
    
    // now initialize v[] using the array of C
    for (int i = 0; i < number; i++)
      (*this)(i) = Element(C(i));
}

Collection::Collection(CONST floatCollection& C) : shape(C.shape)
{ 
    construct(); 

    // now initialize v[] using the array of C
    for (int i = 0; i < number; i++)
      (*this)(i) = Element(C(i));
}

Collection::Collection(CONST doubleCollection& C) : shape(C.shape)
{ 
    construct(); 

    // now initialize v[] using the array of C
    for (int i = 0; i < number; i++)
      (*this)(i) = Element(C(i));
}

Collection::Collection(CONST booleanCollection& C) : shape(C.shape)
{ 
    construct(); 

    // now initialize v[] using the array of C
    for (int i = 0; i < number; i++)
      (*this)(i) = Element(C(i));
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::~Collection: destructor"

Collection::~Collection()
{
    delete v;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_Equals: collection=collection operator"
//  assign the collection C to "this"

void Collection::OP_Equals(CONST Collection &C)
{
    // test for assignment to itself: C = C
    if (this == &C) 
      return;
    
    INVARIANT4(shape == C.shape, // check for conformity
	       "Shape mismatch: %s and %s", shape.name(), C.shape.name());
    
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) = C(i);
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_Equals: collection=scalar operator"
//  assign the integer x to all elements of "this"

void Collection::OP_Equals(CONST Element x)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) = x;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_PlusEquals: collection+=collection operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_PlusEquals(CONST Collection &C)
{
    INVARIANT4(shape == C.shape, // check for conformity
	       "Shape mismatch: %s and %s", shape.name(), C.shape.name());
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) += C(i);
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_PlusEquals: collection+=scalar operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_PlusEquals(CONST Element x)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) += x;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_MinusEquals: collection-=collection operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_MinusEquals(CONST Collection &C)
{
    INVARIANT4(shape == C.shape, // check for conformity
	       "Shape mismatch: %s and %s", shape.name(), C.shape.name());
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) -= C(i);
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_MinusEquals: collection-=scalar operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_MinusEquals(CONST Element x)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) -= x;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_TimesEquals: collection*=collection operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_TimesEquals(CONST Collection &C)
{
    INVARIANT4(shape == C.shape, // check for conformity
	       "Shape mismatch: %s and %s", shape.name(), C.shape.name());
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) *= C(i);
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_TimesEquals: collection*=scalar operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_TimesEquals(CONST Element x)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) *= x;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_DivideEquals: collection/=collection operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_DivideEquals(CONST Collection &C)
{
    INVARIANT4(shape == C.shape, // check for conformity
	       "Shape mismatch: %s and %s", shape.name(), C.shape.name());
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i)) {
	  INVARIANT3(C(i) != 0, "/= division by zero at index %d", i);
	  (*this)(i) /= C(i);
      }
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_DivideEquals: collection/=scalar operator"

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void Collection::OP_DivideEquals(CONST Element x)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    INVARIANT2(x != 0, "/= division by zero scalar");

    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) /= x;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_ModEquals: collection%=collection operator"

#ifdef INT_ELEMENTS		// not for char, float, double, or boolean
void Collection::OP_ModEquals(CONST Collection &C)
{
    INVARIANT4(shape == C.shape, // check for conformity
	       "Shape mismatch: %s and %s", shape.name(), C.shape.name());
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());

    for (int i = 0; i < number; i++)
      if (Active.isActive(i)) {
	  INVARIANT3(C(i) != 0, "%%= mod by zero at index %d", i);
	  (*this)(i) %= C(i);
      }
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_ModEquals: collection%=scalar operator"

#ifdef INT_ELEMENTS		// not for char, float, double, or boolean
void Collection::OP_ModEquals(CONST Element x)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());

    INVARIANT2(x != 0, "%%= mod by zero scalar");

    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	(*this)(i) %= x;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_PlusPlus: ++collection operator"
//   prefix increment operator 

#ifndef BOOLEAN_ELEMENTS			  // booleans don't support ++ and --
void Collection::OP_PlusPlus(void)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());

    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	++((*this)(i));
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_PlusPlus: collection++ operator"
//   postfix increment operator

#ifndef BOOLEAN_ELEMENTS			  // booleans don't support ++ and --
void Collection::OP_PlusPlus(CONST int zero)
{
    INVARIANT3(zero == 0,		// compiler is supposed to pass only zero
	       "Huh? Compiler passed non-zero argument %d to postfix++ operato", zero);
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());

    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	((*this)(i))++;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_MinusMinus: --collection operator"
//   prefix decrement operator

#ifndef BOOLEAN_ELEMENTS			  // booleans don't support ++ and --
void Collection::OP_MinusMinus(void)
{
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());

    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	--((*this)(i));
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "Collection::OP_MinusMinus: collection-- operator"
//   postfix decrement operator

#ifndef BOOLEAN_ELEMENTS			  // booleans don't support ++ and --
void Collection::OP_MinusMinus(CONST int zero)
{
    INVARIANT3(zero == 0,		// compiler is supposed to pass only zero
	       "Huh? Compiler passed non-zero argument %d to postfix-- operato", zero);
    INVARIANT4(Active.shapeOK(shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       shape.name(), Active.shapeName());
    
    for (int i = 0; i < number; i++)
      if (Active.isActive(i))
	((*this)(i))--;
}
#endif

//////////////// Friends of the Collection class ///////////////////

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Minus: -collection operator"
//  the unary minus operator
// (the unary plus operator is not implemented here, as it is trivial).

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Minus(Collection& R, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    OP_Minus(R, (Element) 0, A);	// R = 0-A, ie, subtract it from zero
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Plus: collection+collection operator"
//  add collections A + B

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Plus(Collection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) + B(i);
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Plus: collection+scalar operator"
//  add collection A to integer x

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Plus(Collection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) + x;
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Plus: scalar+collection operator"
//  add collection A to integer x

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Plus(Collection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = x + A(i);
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Minus: collection-collection operator"
//  subtract collections A-B

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Minus(Collection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) - B(i);
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Minus: collection-scalar operator"
//  subtract the integer x from collection A, 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Minus(Collection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) - x;
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Minus: scalar-collection operator"
//  subtract collection A from the integer x, 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Minus(Collection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = x - A(i);
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Times: collection*collection operator"
//  multiply collections A*B (elementwise), 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Times(Collection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) * B(i);
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Times: collection*scalar operator"
//  multiply collection A by integer x, 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Times(Collection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) * x;
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Times: scalar*collection operator"
//  multiply collection A by integer x, 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Times(Collection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = x * A(i);
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Divide: collection/collection operator"
//  divide collection A by collection B (elementwise), 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Divide(Collection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i)) {
	  INVARIANT3(B(i) != 0, "/ division by zero at index %d", i);
	  R(i) = A(i) / B(i);
      } else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Divide: collection/scalar operator"
//  divide collection A by integer x (elementwise), 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Divide(Collection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    INVARIANT2(x != 0, "/ division by zero scalar");

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) / x;
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Divide: scalar/collection operator"
//  divide integer x by collection A (elementwise), 

#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
void OP_Divide(Collection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i)) {
	  INVARIANT3(A(i) != 0, "/ division by zero at index %d", i);
	  R(i) = x / A(i);
      } else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Mod: collection%collection operator"
//  mod collection A by collection B (elementwise), 

#ifdef INT_ELEMENTS		// not for char, float, double, or boolean
void OP_Mod(Collection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i)) {
	  INVARIANT3(B(i) != 0, "%% mod by zero at index %d", i);
	  R(i) = A(i) % B(i);
      } else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Mod: collection%scalar operator"
//  mod collection A by integer x (elementwise), 

#ifdef INT_ELEMENTS		// not for char, float, double, or boolean
void OP_Mod(Collection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    INVARIANT2(x != 0, "%% mod by zero scalar");

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = A(i) % x;
      else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_Mod: scalar%collection operator"
//  mod integer x by collection A (elementwise), 

#ifdef INT_ELEMENTS		// not for char, float, double, or boolean
void OP_Mod(Collection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i)) {
	  INVARIANT3(A(i) != 0, "%% mod by zero at index %d", i);
	  R(i) = x % A(i);
      } else 
	R(i) = 0;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_LT: collection<collection operator"
// R = collection A < collection B (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_LT(booleanCollection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) < B(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_LT: collection<scalar operator"
// R = collection A < integer x (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_LT(booleanCollection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) < x);
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_LT: scalar<collection operator"
// R = integer x < collection A (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_LT(booleanCollection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (x < A(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_GT: collection>collection operator"
// R = collection A > collection B (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_GT(booleanCollection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) > B(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_GT: collection>scalar operator"
// R = collection A > integer x (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_GT(booleanCollection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) > x);
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_GT: scalar>collection operator"
// R = integer x > collection A (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_GT(booleanCollection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (x > A(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_LE: collection<=collection operator"
// R = collection A <= collection B (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_LE(booleanCollection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) <= B(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_LE: collection<=scalar operator"
// R = collection A <= integer x (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_LE(booleanCollection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) <= x);
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_LE: scalar<=collection operator"
// R = integer x <= collection A (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_LE(booleanCollection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (x <= A(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_GE: collection>=collection operator"
// R = collection A >= collection B (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_GE(booleanCollection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) >= B(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_GE: collection>=scalar operator"
// R = collection A >= integer x (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_GE(booleanCollection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) >= x);
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_GE: scalar>=collection operator"
// R = integer x >= collection A (elementwise), 

#ifndef BOOLEAN_ELEMENTS	// don't allow comparisons on booleans
void OP_GE(booleanCollection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (x >= A(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_EQ: collection==collection operator"
// R = collection A == collection B (elementwise), 

void OP_EQ(booleanCollection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) == B(i));
      else 
	R(i) = false;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_EQ: collection==scalar operator"
// R = collection A == integer x (elementwise), 

void OP_EQ(booleanCollection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) == x);
      else 
	R(i) = false;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_EQ: scalar==collection operator"
// R = integer x == collection A (elementwise), 

void OP_EQ(booleanCollection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (x == A(i));
      else 
	R(i) = false;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_NE: collection!=collection operator"
// R = collection A != collection B (elementwise), 

void OP_NE(booleanCollection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) != B(i));
      else 
	R(i) = false;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_NE: collection!=scalar operator"
// R = collection A != integer x (elementwise), 

void OP_NE(booleanCollection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) != x);
      else 
	R(i) = false;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_NE: scalar!=collection operator"
// R = integer x != collection A (elementwise), 

void OP_NE(booleanCollection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
      R(i) = (x != A(i));
      else 
	R(i) = false;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_AND: collection&&collection operator"
// R = collection A && collection B (elementwise), 

#ifdef BOOLEAN_ELEMENTS		// only allowed for booleans
void OP_AND(Collection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) && B(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_AND: collection&&scalar operator"
// R = collection A && integer x (elementwise)

#ifdef BOOLEAN_ELEMENTS		// only allowed for booleans
void OP_AND(Collection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) && x);
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_AND: scalar&&collection operator"
// R = integer x && collection A (elementwise),

#ifdef BOOLEAN_ELEMENTS		// only allowed for booleans
void OP_AND(Collection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (x && A(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_OR: collection||collection operator"
// R = collection A || collection B (elementwise),

#ifdef BOOLEAN_ELEMENTS		// only allowed for booleans
void OP_OR(Collection& R, CONST Collection &A, CONST Collection &B)
{
    INVARIANT4(A.shape == B.shape, // check for conformity
	       "Shape mismatch: %s and %s", A.shape.name(), B.shape.name());
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (A(i) || B(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_OR: collection||scalar operator"
// R = collection A || integer x (elementwise)

#ifdef BOOLEAN_ELEMENTS		// only allowed for booleans
void OP_OR(Collection& R, CONST Collection &A, CONST Element x)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = ( A(i) || x);
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_OR: scalar||collection operator"
// R = integer x || collection A (elementwise)

#ifdef BOOLEAN_ELEMENTS		// only allowed for booleans
void OP_OR(Collection& R, CONST Element x, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (x || A(i));
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "OP_NOT: !collection operator"
// R = collection !A (elementwise)

#ifdef BOOLEAN_ELEMENTS		// only allowed for booleans
void OP_NOT(Collection& R, CONST Collection &A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = !A(i);
      else 
	R(i) = false;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "any(Collection)"
//  apply an or-reduction to collection A, returning a boolean result

boolean any(CONST Collection& A)
{
    boolean any = false;
    
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	any = (any || (A(i) != 0));
    
    return any;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "all(Collection)"
//  apply an and-reduction to collection A, returning a boolean result

boolean all(CONST Collection& A)
{
    boolean all = true;
    
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	all = (all && (A(i) != 0));
    
    return all;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "n_nonzeros(Collection)"
// count how many non-zeros are in collection A, returning the number

int n_nonzeros(CONST Collection& A)
{
    int n = 0;
    
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	if (A(i) != 0) n++;
    
    return n;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "sum(Collection)"
//  apply a sum-reduction to collection A, returning the sum

#ifndef BOOLEAN_ELEMENTS			  // booleans don't support sum, max, or min
Element sum(CONST Collection& A)
{
    Element sum = 0;
    
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++) 
      if (Active.isActive(i))
	sum += A(i);
    
    return sum;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "max_value(Collection)"
//  apply a max-reduction to collection A, returning the max
// if the active set is empty, it returns the smallest number of this type.

#ifndef BOOLEAN_ELEMENTS			  // booleans don't support max and min
Element max_value(CONST Collection& A)
{
    Element m = Smallest();
    
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	m = max(m, A(i));
    
    return m;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "min_value(Collection)"
//  apply a min-reduction to collection A, returning the min
// if the active set is empty, it returns the largest number of this type.

#ifndef BOOLEAN_ELEMENTS			  // booleans don't support max and min
Element min_value(CONST Collection& A)
{
    Element m = Biggest();
    
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	m = min(m, A(i));
    
    return m;
}
#endif

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "n_active(Collection)"
// count how many active processors there are, returning the number
// note this is equivalent to ((Collection)1).n_nonzeros but much clearer.

int n_active(CONST Collection& A) 
{
    int n = 0;
    
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	n++;
    
    return n;
}

/////////////////////////////////////////////////////////////////////
// @SUBTITLE "apply(function,Collection): elementwise function-apply"
//  apply the function f to each element of the collection
//  the function should not be allowed to have side effects
//  f is a pointer to function taking Element, returning Element

void apply(Collection& R, Element (*f)(CONST Element), CONST Collection& A)
{
    INVARIANT4(Active.shapeOK(A.shape), // check conformity with active set
	       "Operand shape %s does not match context shape %s",
	       A.shape.name(), Active.shapeName());

    for (int i = 0; i < A.number; i++)
      if (Active.isActive(i))
	R(i) = (*f)(A(i));
      else
	R(i) = false;
}

/////////////////////////////////////////////////////////////////////
