// @TITLE "Collection.h: class Collection"
//
// included by dapple.h -- do not include this directly!
//
// Collection class is the base class for collections of 
// different shapes, eg, Vector and Matrix.  This class supports only
// elementwise operators, and reductions.  Anything that assumes some
// order on the elements of a collection, such as shift, rotate, permute,
// scans, input, and output, must be implemented in the index space 
// of the derived class. 
//
// Most methods are defined as friend functions rather than as member 
// functions, as suggested in ARM p. 249, because then it can be explicit
// that the Collection parameter
// need not be an lvalue (ie, it is const), and because it leads to 
// an interface that looks functional.  The latter doesn't make much difference
// here, but once we get to Vectors, A = plus_scan(B) makes more sense than 
// A = B.plus_scan(), especially because one might think that B.plus_scan()
// by itself makes sense.
// 
// David Kotz 1994

// $Id: Collection.h,v 2.16 95/12/17 00:07:44 dfk release Locker: dfk $

class intCollection;
class charCollection;
class floatCollection;
class doubleCollection;
class booleanCollection;

class Collection {

    friend class ActiveSet;

    // all collections are mutual friends; 
    // this allows the conversion constructors to work
#if !defined(INT_ELEMENTS)
    friend class intCollection;
    friend class intMatrix;
#endif
#if !defined(CHAR_ELEMENTS)
    friend class charCollection;
    friend class charMatrix;
#endif
#if !defined(FLOAT_ELEMENTS)
    friend class floatCollection;
    friend class floatMatrix;
#endif
#if !defined(DOUBLE_ELEMENTS)
    friend class doubleCollection;
    friend class doubleMatrix;
#endif
#if !defined(BOOLEAN_ELEMENTS)
    friend class booleanCollection;
    friend class booleanMatrix;
#endif

protected:			// only usable by subclasses and friends
    // constructors and destructors
    Collection(CONST Shape& s)           : shape(s) { construct(); }
    Collection(CONST Shape& s, CONST Element x): shape(s) { construct(); init(x); }
    Collection(CONST intCollection& C);	// similar to the above
    Collection(CONST charCollection& C); // but defined in Collection.cc
    Collection(CONST floatCollection& C);
    Collection(CONST doubleCollection& C);
    Collection(CONST booleanCollection& C);
    virtual ~Collection();	// virtual in case derived classes have a destructor

public:				// inherited by all subclasses
    // reductions
    friend boolean any(CONST Collection&);	  // do an  or-reduction 
    friend boolean all(CONST Collection&);	  // do an and-reduction 
    friend int     n_nonzeros(CONST Collection&); // count the nonzeros
#ifndef BOOLEAN_ELEMENTS			  // booleans don't support sum, max, or min
    friend Element sum(CONST Collection&);        // do a  sum-reduction
    friend Element max_value(CONST Collection&);  // do a max-reduction
    friend Element min_value(CONST Collection&);  // do a min-reduction
#endif
    friend int        n_active(CONST Collection&);   // # VPs that are active

protected:			// used by subclasses to execute elementwise operators
    void OP_Equals(CONST Collection& A);       // this = A
    void OP_Equals(CONST Element x);	       // this = x
#ifndef BOOLEAN_ELEMENTS		       // don't allow arithmetic on booleans
    void OP_PlusEquals(CONST Collection& A);   // this += A
    void OP_PlusEquals(CONST Element x);       // this += x
    void OP_MinusEquals(CONST Collection& A);  // this -= A
    void OP_MinusEquals(CONST Element x);      // this -= x
    void OP_TimesEquals(CONST Collection& A);  // this *= A
    void OP_TimesEquals(CONST Element x);      // this *= x
    void OP_DivideEquals(CONST Collection& A); // this /= A
    void OP_DivideEquals(CONST Element x);     // this /= x
# ifdef INT_ELEMENTS			       // not for char, float, double, or boolean
    void OP_ModEquals(CONST Collection& A);    // this %= A
    void OP_ModEquals(CONST Element x);	       // this %= x
# endif
    void OP_PlusPlus(void);	              // ++this
    void OP_PlusPlus(CONST int zero);	      // this++
    void OP_MinusMinus(void);		      // --this
    void OP_MinusMinus(CONST int zero);	      // this--
//  friend void OP_Plus(Collection& R, CONST Collection& A);  // R = +A (not needed)
    friend void OP_Minus(Collection& R, CONST Collection& A); // R = -A 
    friend void OP_Plus(Collection& R, CONST Collection& A, CONST Collection& B);   // R = A + B
    friend void OP_Plus(Collection& R, CONST Collection& A, CONST Element x);	    // R = A + x
    friend void OP_Plus(Collection& R, CONST Element x, CONST Collection& B);	    // R = x + A
    friend void OP_Minus(Collection& R, CONST Collection& A, CONST Collection& B);  // R = A - B
    friend void OP_Minus(Collection& R, CONST Collection& A, CONST Element x);	    // R = A - x
    friend void OP_Minus(Collection& R, CONST Element x, CONST Collection& B);	    // R = x - B
    friend void OP_Times(Collection& R, CONST Collection& A, CONST Collection& B);  // R = A * B
    friend void OP_Times(Collection& R, CONST Collection& A, CONST Element x);	    // R = A * x
    friend void OP_Times(Collection& R, CONST Element x, CONST Collection& B);	    // R = x * B
    friend void OP_Divide(Collection& R, CONST Collection& A, CONST Collection& B); // R = A / B
    friend void OP_Divide(Collection& R, CONST Collection& A, CONST Element x);	    // R = A / x
    friend void OP_Divide(Collection& R, CONST Element x, CONST Collection& B);	    // R = x / B
# ifdef INT_ELEMENTS			       // not for char, float, double, or boolean
    friend void OP_Mod(Collection& R, CONST Collection& A, CONST Collection& B);    // R = A % B
    friend void OP_Mod(Collection& R, CONST Collection& A, CONST Element x);	    // R = A % x
    friend void OP_Mod(Collection& R, CONST Element x, CONST Collection& B);	    // R = x % B
# endif // INT_ELEMENTS

    friend void OP_LT(booleanCollection& R, CONST Collection& A, CONST Collection& B); // R = A < B
    friend void OP_LT(booleanCollection& R, CONST Collection& A, CONST Element x);     // R = A < x
    friend void OP_LT(booleanCollection& R, CONST Element x, CONST Collection& B);     // R = x < B
    friend void OP_GT(booleanCollection& R, CONST Collection& A, CONST Collection& B); // R = A > B
    friend void OP_GT(booleanCollection& R, CONST Collection& A, CONST Element x);     // R = A > x
    friend void OP_GT(booleanCollection& R, CONST Element x, CONST Collection& B);     // R = x > B
    friend void OP_LE(booleanCollection& R, CONST Collection& A, CONST Collection& B); // R = A <= B
    friend void OP_LE(booleanCollection& R, CONST Collection& A, CONST Element x);     // R = a <= x
    friend void OP_LE(booleanCollection& R, CONST Element x, CONST Collection& B);     // R = x <= B
    friend void OP_GE(booleanCollection& R, CONST Collection& A, CONST Collection& B); // R = A >= B
    friend void OP_GE(booleanCollection& R, CONST Collection& A, CONST Element x);     // R = A >= x
    friend void OP_GE(booleanCollection& R, CONST Element x, CONST Collection& B);     // R = x >= B
#endif // BOOLEAN_ELEMENTS
    friend void OP_EQ(booleanCollection& R, CONST Collection& A, CONST Collection& B); // R = A == B
    friend void OP_EQ(booleanCollection& R, CONST Collection& A, CONST Element x);     // R = A == x
    friend void OP_EQ(booleanCollection& R, CONST Element x, CONST Collection& B);     // R = x == B
    friend void OP_NE(booleanCollection& R, CONST Collection& A, CONST Collection& B); // R = A != B
    friend void OP_NE(booleanCollection& R, CONST Collection& A, CONST Element x);     // R = A != x
    friend void OP_NE(booleanCollection& R, CONST Element x, CONST Collection& B);     // R = x != B

#ifdef BOOLEAN_ELEMENTS   // only for booleans
    friend void OP_AND(booleanCollection& R, CONST Collection& A, CONST Collection& B); // R = A && B
    friend void OP_AND(booleanCollection& R, CONST Collection& A, CONST Element x);	// R = A && x
    friend void OP_AND(booleanCollection& R, CONST Element x, CONST Collection& B);	// R = x && B
    friend void OP_OR(booleanCollection& R, CONST Collection& A, CONST Collection& B);  // R = A || B
    friend void OP_OR(booleanCollection& R, CONST Collection& A, CONST Element x);	// R = A || x
    friend void OP_OR(booleanCollection& R, CONST Element x, CONST Collection& B);	// R = x || B
    friend void OP_NOT(booleanCollection& R, CONST Collection& A);			// R = !A
#endif // BOOLEAN_ELEMENTS
    
    // misc
    friend void apply(Collection& R, Element (*f)(CONST Element), CONST Collection& V); // R = f(V)

// INTERNAL but needs to be public because of complex friendship relationship.
public:
    Shape shape;		// the shape of this object
    int number;			// cached result of shape.number()
    Element *v;			// essentially an array of [number] elements

    // this is how we reference the elements of the collection, internally:
    // we can refer to element i of any collection C as C(i)
    Element& operator()(CONST int i) CONST
      { return v[i]; } 
    // we can refer to element in row r, column c of matrix M as M(r,c)
    Element& operator()(CONST int r, CONST int c) CONST
      { return v[r * shape.dim2 + c]; } 

private: 
    // set up the collection, but do not initialize it.
    void construct(void);

    // initialize from a scalar
    void init(CONST Element x);
};
