// @TITLE "Slice.h: class Slice"
//
// included by dapple.h -- do not include this directly!
//
// Slice class.  A slice is created as a row or column
// from a matrix.  A Slice is a subclass of Vector, so this behaves like
// a vector for most operations.  However, we overload the assignment 
// operators, so that we can provide the illusion that this is really 
// a reference to the original matrix.  Thus, you can write expressions
// like
//        A.row(i) -= pivotValue * A.row(pivotRow);
//
// David Kotz 1994

// $Id: Slice.h,v 1.14 95/12/17 00:07:54 dfk release Locker: dfk $

#ifndef SLICE_H			// once-only stuff (note, not the whole file!!)
#define SLICE_H

enum SliceType {rowSlice, colSlice};

class SliceDummy { 	// a placeholder used to make column slices
public:
    SliceDummy(void) { x = 0; } 
    int x; 
};

extern CONST SliceDummy _;	// yes, that variable name is underscore '_'

#endif // SLICE_H			// end once-only stuff

class Matrix;

// @SUBTITLE "class Slice"
class Slice {
public: 
    // make a new slice
    Slice(CONST SliceType t, CONST int index, Matrix *CONST m) 
      : type(t), rowcol(index), M(m) {};
    
    // conversion operator to turn a slice into a vector, if needed
    operator Vector () CONST;
    
    // second-subscripting a matrix means subscripting a slice
    // M[i][j] is an Element reference
    Element& operator[](CONST int index) CONST;
    // M[i][_] is the same as M[i]
    Slice operator[](CONST SliceDummy&) CONST;

    // overload the assignment operators:
    Slice& operator=(CONST Vector& B);   // A=B
    Slice& operator=(CONST Element x);	 // A=x
#ifndef BOOLEAN_ELEMENTS // no arithmetic on booleans
    Slice& operator+=(CONST Vector& B);  // A+=B
    Slice& operator+=(CONST Element x);	 // A+=x
    Slice& operator-=(CONST Vector& B);  // A-=B
    Slice& operator-=(CONST Element x);	 // A-=x
    Slice& operator*=(CONST Vector& B);  // A*=B
    Slice& operator*=(CONST Element x);	 // A*=x
    Slice& operator/=(CONST Vector& B);  // A/=B
    Slice& operator/=(CONST Element x);	 // A/=x
# ifdef INT_ELEMENTS		// not for char, float, double, or boolean
    Slice& operator%=(CONST Vector& B);  // A%=B
    Slice& operator%=(CONST Element x);	 // A%=x
# endif // INT_ELEMENTS
    Slice  operator++(void);	         // ++A
    Slice  operator++(int zero);	 // A++
    Slice  operator--(void);		 // --A
    Slice  operator--(int zero);	 // A--
#endif // BOOLEAN_ELEMENTS

private: 
    CONST SliceType type;	// row or column?
    CONST int rowcol;		// which one?
    Matrix *CONST M;		// remember our heritage
};

// @SUBTITLE "class HalfSlice"
// A half slice is created with M[_]..., ie, it is an unspecified column slice.
// Thus, the only valid operation is to subscript it, M[_][j] into a full Slice

class HalfSlice {
public: 
    // make a new half-slice
    HalfSlice(Matrix *CONST m) : M(m) {}
    
    // M[_][j] is a column slice
    Slice operator[](CONST int index) CONST;

private:
    Matrix *CONST M;		// remember our heritage
};

